@@ -75,6 +75,7 @@ gid_t GdmGroupId;		/* Groupid under which gdm should run */
 pid_t extra_process = -1;	/* An extra process.  Used for quickie 
 				   processes, so that they also get whacked */
 int extra_status = 0;		/* Last status from the last extra process */
+pid_t gdm_main_pid = 0;
 gboolean preserve_ld_vars = FALSE; /* Preserve the ld environment variables */
 gboolean no_daemon = FALSE;	/* Do not daemonize */
 
@@ -101,6 +102,7 @@ gchar *GdmLocaleFile = NULL;
 gchar *GdmGnomeDefaultSession = NULL;
 gchar *GdmAutomaticLogin = NULL;
 gboolean GdmAutomaticLoginEnable = FALSE;
+gchar *GdmLocalNoPasswordUsers = NULL;
 gboolean GdmAlwaysRestartServer = FALSE;
 gchar *GdmConfigurator = NULL;
 gboolean GdmConfigAvailable = FALSE;
@@ -119,8 +121,11 @@ gchar *GdmPostSession = NULL;
 gchar *GdmFailsafeXServer = NULL;
 gchar *GdmXKeepsCrashing = NULL;
 gchar *GdmHalt = NULL;
+gchar *GdmHaltReal = NULL;
 gchar *GdmReboot = NULL;
+gchar *GdmRebootReal = NULL;
 gchar *GdmSuspend = NULL;
+gchar *GdmSuspendReal = NULL;
 gchar *GdmServAuthDir = NULL;
 gchar *GdmUserAuthDir = NULL;
 gchar *GdmUserAuthFile = NULL;
@@ -227,6 +232,7 @@ gdm_config_parse (void)
     GdmDisplayInit = gnome_config_get_string (GDM_KEY_INITDIR);
     GdmAutomaticLoginEnable = gnome_config_get_bool (GDM_KEY_AUTOMATICLOGIN_ENABLE);
     GdmAutomaticLogin = gnome_config_get_string (GDM_KEY_AUTOMATICLOGIN);
+    GdmLocalNoPasswordUsers = gnome_config_get_string (GDM_KEY_LOCALNOPASSWORDUSERS);
     GdmAlwaysRestartServer = gnome_config_get_bool (GDM_KEY_ALWAYSRESTARTSERVER);
     GdmGreeter = gnome_config_get_string (GDM_KEY_GREETER);
     GdmRemoteGreeter = gnome_config_get_string (GDM_KEY_REMOTEGREETER);
@@ -579,6 +585,11 @@ gdm_config_parse (void)
 	    gdm_fail (_("%s: The gdm group should not be root. Aborting!"), "gdm_config_parse");
     }
 
+    /* get the actual commands to use */
+    GdmHaltReal = ve_get_first_working_command (GdmHalt, FALSE);
+    GdmRebootReal = ve_get_first_working_command (GdmReboot, FALSE);
+    GdmSuspendReal = ve_get_first_working_command (GdmSuspend, FALSE);
+
     setegid (GdmGroupId);	/* gid remains `gdm' */
     seteuid (GdmUserId);
 
@@ -711,8 +722,8 @@ gdm_daemonify (void)
     pid = fork ();
     if (pid > 0) {
 
-	if ((pf = fopen (GdmPidFile, "w"))) {
-	    fprintf (pf, "%d\n", pid);
+	if ((pf = fopen (GdmPidFile, "w")) != NULL) {
+	    fprintf (pf, "%d\n", (int)pid);
 	    fclose (pf);
 	}
 
@@ -733,9 +744,9 @@ gdm_daemonify (void)
     close (1);
     close (2);
 
-    open("/dev/null", O_RDONLY); /* open stdin - fd 0 */
-    open("/dev/null", O_RDWR); /* open stdout - fd 1 */
-    open("/dev/null", O_RDWR); /* open stderr - fd 2 */
+    gdm_open_dev_null (O_RDONLY); /* open stdin - fd 0 */
+    gdm_open_dev_null (O_RDWR); /* open stdout - fd 1 */
+    gdm_open_dev_null (O_RDWR); /* open stderr - fd 2 */
 }
 
 static void 
@@ -774,13 +785,13 @@ gdm_start_first_unborn_local (int delay)
 	}
 }
 
-static void
-final_cleanup (void)
+void
+gdm_final_cleanup (void)
 {
-	GSList *list;
+	GSList *list, *li;
 	sigset_t mask;
 
-	gdm_debug ("final_cleanup");
+	gdm_debug ("gdm_final_cleanup");
 
 	sigemptyset (&mask);
 	sigaddset (&mask, SIGCHLD);
@@ -794,7 +805,30 @@ final_cleanup (void)
 	}
 
 	list = g_slist_copy (displays);
-	g_slist_foreach (list, (GFunc) gdm_display_unmanage, NULL);
+	for (li = list; li != NULL; li = li->next) {
+		GdmDisplay *d = li->data;
+		/* XDMCP and FLEXI_XNEST are safe to kill
+		 * immediately */
+		if (d->type == TYPE_XDMCP ||
+		    d->type == TYPE_FLEXI_XNEST)
+			gdm_display_unmanage (d);
+	}
+	g_slist_free (list);
+
+	list = g_slist_copy (displays);
+	/* somewhat of a hack to kill last server
+	 * started first.  This mostly makes things end up on
+	 * the right vt */
+	list = g_slist_reverse (list);
+	for (li = list; li != NULL; li = li->next) {
+		GdmDisplay *d = li->data;
+		/* HACK! Wait 2 seconds between killing of local servers
+		 * because X is stupid and full of races and will hang my
+		 * keyboard if I don't */
+		if (li != list)
+			sleep (2);
+		gdm_display_unmanage (d);
+	}
 	g_slist_free (list);
 
 	/* Close stuff */
@@ -864,9 +898,9 @@ deal_with_x_crashes (GdmDisplay *d)
 
 		    /* No error checking here - if it's messed the best response
 		    * is to ignore & try to continue */
-		    open ("/dev/null", O_RDONLY); /* open stdin - fd 0 */
-		    open ("/dev/null", O_RDWR); /* open stdout - fd 1 */
-		    open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
+		    gdm_open_dev_null (O_RDONLY); /* open stdin - fd 0 */
+		    gdm_open_dev_null (O_RDWR); /* open stdout - fd 1 */
+		    gdm_open_dev_null (O_RDWR); /* open stderr - fd 2 */
 
 		    argv[0] = GdmXKeepsCrashing;
 		    argv[1] = NULL;
@@ -939,26 +973,6 @@ deal_with_x_crashes (GdmDisplay *d)
     return FALSE;
 }
 
-static gboolean
-bin_executable (const char *command)
-{
-	char **argv;
-
-	if (ve_string_empty (command))
-		return FALSE;
-
-	argv = ve_split (command);
-	if (argv != NULL &&
-	    argv[0] != NULL &&
-	    access (argv[0], X_OK) == 0) {
-		g_strfreev (argv);
-		return TRUE;
-	} else {
-		g_strfreev (argv);
-		return FALSE;
-	}
-}
-
 static gboolean 
 gdm_cleanup_children (void)
 {
@@ -1004,22 +1018,22 @@ gdm_cleanup_children (void)
 
 	    if (d->sesspid > 0) {
 		    if (kill (d->sesspid, SIGTERM) == 0)
-			    waitpid (d->sesspid, NULL, 0);
+			    ve_waitpid_no_signal (d->sesspid, NULL, 0);
 		    d->sesspid = 0;
 	    }
 	    if (d->greetpid > 0) {
 		    if (kill (d->greetpid, SIGTERM) == 0)
-			    waitpid (d->greetpid, NULL, 0);
+			    ve_waitpid_no_signal (d->greetpid, NULL, 0);
 		    d->greetpid = 0;
 	    }
 	    if (d->chooserpid > 0) {
 		    if (kill (d->chooserpid, SIGTERM) == 0)
-			    waitpid (d->chooserpid, NULL, 0);
+			    ve_waitpid_no_signal (d->chooserpid, NULL, 0);
 		    d->chooserpid = 0;
 	    }
 	    if (d->servpid > 0) {
 		    if (kill (d->servpid, SIGTERM) == 0)
-			    waitpid (d->servpid, NULL, 0);
+			    ve_waitpid_no_signal (d->servpid, NULL, 0);
 		    d->servpid = 0;
 	    }
     }
@@ -1063,21 +1077,23 @@ gdm_cleanup_children (void)
     /* checkout if we can actually do stuff */
     switch (status) {
     case DISPLAY_REBOOT:
-	    if ( ! bin_executable (GdmReboot))
+	    if (GdmRebootReal == NULL)
 		    status = DISPLAY_REMANAGE;
 	    break;
     case DISPLAY_HALT:
-	    if ( ! bin_executable (GdmHalt))
+	    if (GdmHaltReal == NULL)
 		    status = DISPLAY_REMANAGE;
 	    break;
     case DISPLAY_SUSPEND:
-	    if ( ! bin_executable (GdmSuspend))
+	    if (GdmSuspendReal == NULL)
 		    status = DISPLAY_REMANAGE;
 	    break;
     default:
 	    break;
     }
 
+start_autopsy:
+
     /* Autopsy */
     switch (status) {
 	
@@ -1096,34 +1112,46 @@ gdm_cleanup_children (void)
     case DISPLAY_REBOOT:	/* Reboot machine */
 	gdm_info (_("gdm_child_action: Master rebooting..."));
 
-	final_cleanup ();
+	gdm_final_cleanup ();
+	chdir ("/");
 
-	argv = ve_split (GdmReboot);
+	argv = ve_split (GdmRebootReal);
 	execv (argv[0], argv);
 
 	gdm_error (_("gdm_child_action: Reboot failed: %s"), strerror (errno));
+
+	status = DISPLAY_REMANAGE;
+	goto start_autopsy;
 	break;
 	
     case DISPLAY_HALT:		/* Halt machine */
 	gdm_info (_("gdm_child_action: Master halting..."));
 
-	final_cleanup ();
+	gdm_final_cleanup ();
+	chdir ("/");
 
-	argv = ve_split (GdmHalt);
+	argv = ve_split (GdmHaltReal);
 	execv (argv[0], argv);
 
 	gdm_error (_("gdm_child_action: Halt failed: %s"), strerror (errno));
+
+	status = DISPLAY_REMANAGE;
+	goto start_autopsy;
 	break;
 
     case DISPLAY_SUSPEND:	/* Suspend machine */
 	gdm_info (_("gdm_child_action: Master suspending..."));
 
-	final_cleanup ();
+	gdm_final_cleanup ();
+	chdir ("/");
 
-	argv = ve_split (GdmSuspend);
+	argv = ve_split (GdmSuspendReal);
 	execv (argv[0], argv);
 
 	gdm_error (_("gdm_child_action: Suspend failed: %s"), strerror (errno));
+
+	status = DISPLAY_REMANAGE;
+	goto start_autopsy;
 	break;
 
     case DISPLAY_RESTARTGDM:
@@ -1219,7 +1247,7 @@ static void
 gdm_restart_now (void)
 {
 	gdm_info (_("GDM restarting ..."));
-	final_cleanup ();
+	gdm_final_cleanup ();
 	if (stored_path != NULL)
 		gnome_setenv ("PATH", stored_path, TRUE);
 	execvp (stored_argv[0], stored_argv);
@@ -1243,7 +1271,7 @@ gdm_safe_restart (void)
 }
 
 static gboolean
-mainloop_sig_callback (gint8 sig, gpointer data)
+mainloop_sig_callback (int sig, gpointer data)
 {
   gdm_debug ("mainloop_sig_callback: Got signal %d", (int)sig);
   switch (sig)
@@ -1256,7 +1284,7 @@ mainloop_sig_callback (gint8 sig, gpointer data)
     case SIGINT:
     case SIGTERM:
       gdm_debug ("mainloop_sig_callback: Got TERM/INT. Going down!");
-      final_cleanup ();
+      gdm_final_cleanup ();
       exit (EXIT_SUCCESS);
       break;
 
@@ -1276,12 +1304,6 @@ mainloop_sig_callback (gint8 sig, gpointer data)
   return TRUE;
 }
 
-static void
-signal_notify (int sig)
-{
-	gdm_signal_notify (sig);
-}
-
 /*
  * main: The main daemon control
  */
@@ -1344,13 +1366,13 @@ linux_only_is_running (pid_t pid)
 	if (realpath ("/proc/self/exe", resolved_self) == NULL) {
 		g_free (running);
 		/* probably not a linux system */
-		return FALSE;
+		return TRUE;
 	}
 
 	if (realpath (running, resolved_running) == NULL) {
 		g_free (running);
 		/* probably not a linux system */
-		return FALSE;
+		return TRUE;
 	}
 
 	g_free (running);
@@ -1368,12 +1390,7 @@ ensure_desc_012 (void)
 	 * we of course count on the fact that open
 	 * opens the lowest available descriptor */
 	for (;;) {
-		fd = open("/dev/null", O_RDWR);
-		/* what to do on fail?  I dunno,
-		 * just keep on going is the safest
-		 * bet I suppose */
-		if (fd < 0)
-			break;
+		fd = gdm_open_dev_null (O_RDWR);
 		/* Once we are up to 3, we're beyond stdin,
 		 * stdout and stderr */
 		if (fd >= 3) {
@@ -1392,6 +1409,8 @@ main (int argc, char *argv[])
     poptContext ctx;
     int nextopt;
 
+    gdm_main_pid = getpid ();
+
     /* We here ensure descriptors 0, 1 and 2 */
     ensure_desc_012 ();
 
@@ -1447,6 +1466,7 @@ main (int argc, char *argv[])
 	    linux_only_is_running (pidv)) {
 		/* make sure the pid file doesn't get wiped */
 		GdmPidFile = NULL;
+		fclose (pf);
 		gdm_fail (_("gdm already running. Aborting!"));
 	}
 
@@ -1458,8 +1478,8 @@ main (int argc, char *argv[])
     if (no_daemon || getppid() == 1) {
 
 	/* Write pid to pidfile */
-	if ((pf = fopen (GdmPidFile, "w"))) {
-	    fprintf (pf, "%d\n", getpid());
+	if ((pf = fopen (GdmPidFile, "w")) != NULL) {
+	    fprintf (pf, "%d\n", (int)getpid());
 	    fclose (pf);
 	}
     }
@@ -1467,13 +1487,13 @@ main (int argc, char *argv[])
 	gdm_daemonify();
 
     /* Signal handling */
-    gdm_signal_add (SIGCHLD, mainloop_sig_callback, NULL);
-    gdm_signal_add (SIGTERM, mainloop_sig_callback, NULL);
-    gdm_signal_add (SIGINT, mainloop_sig_callback, NULL);
-    gdm_signal_add (SIGHUP, mainloop_sig_callback, NULL);
-    gdm_signal_add (SIGUSR1, mainloop_sig_callback, NULL);
+    ve_signal_add (SIGCHLD, mainloop_sig_callback, NULL);
+    ve_signal_add (SIGTERM, mainloop_sig_callback, NULL);
+    ve_signal_add (SIGINT, mainloop_sig_callback, NULL);
+    ve_signal_add (SIGHUP, mainloop_sig_callback, NULL);
+    ve_signal_add (SIGUSR1, mainloop_sig_callback, NULL);
     
-    term.sa_handler = signal_notify;
+    term.sa_handler = ve_signal_notify;
     term.sa_flags = SA_RESTART;
     sigemptyset (&term.sa_mask);
 
@@ -1493,7 +1513,7 @@ main (int argc, char *argv[])
 	gdm_fail (_("%s: Error setting up USR1 signal handler"),
 		  "gdm_main");
 
-    child.sa_handler = signal_notify;
+    child.sa_handler = ve_signal_notify;
     child.sa_flags = SA_RESTART|SA_NOCLDSTOP;
     sigemptyset (&child.sa_mask);
     sigaddset (&child.sa_mask, SIGCHLD);
@@ -1539,100 +1559,75 @@ main (int argc, char *argv[])
     return EXIT_SUCCESS;	/* Not reached */
 }
 
-/* signal main loop support */
-
-
-typedef struct _SignalSource SignalSource;
-struct _SignalSource {
-	GSource     source;
-
-	gint8       signal;
-	guint8      index;
-	guint8      shift;
-};
-
-static	guint32	signals_notified[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
-
 static gboolean
-gdm_signal_prepare (GSource  *source,
-		    gint     *timeout)
+order_exists (int order)
 {
-	SignalSource *ss = (SignalSource *)source;
-
-	return signals_notified[ss->index] & (1 << ss->shift);
+	GSList *li;
+	for (li = displays; li != NULL; li = li->next) {
+		GdmDisplay *d = li->data;
+		if (d->x_servers_order == order)
+			return TRUE;
+	}
+	return FALSE;
 }
 
-static gboolean
-gdm_signal_check (GSource *source)
+static int
+get_new_order (GdmDisplay *d)
 {
-	SignalSource *ss = (SignalSource *)source;
-
-	return signals_notified[ss->index] & (1 << ss->shift);
+	int order;
+	GSList *li;
+	/* first try the position in the 'displays' list as
+	 * our order */
+	for (order = 0, li = displays; li != NULL; order++, li = li->next) {
+		if (li->data == d)
+			break;
+	}
+	/* next make sure it's unique */
+	while (order_exists (order))
+		order++;
+	return order;
 }
 
-static gboolean
-gdm_signal_dispatch (GSource     *source,
-		     GSourceFunc  callback,
-		     gpointer     user_data)
+static void
+write_x_servers (GdmDisplay *d)
 {
-	SignalSource *ss = (SignalSource *)source;
+	FILE *fp;
+	char *file = g_strconcat (GdmServAuthDir, "/", d->name, ".Xservers", NULL);
+	int i;
+	int bogusname;
 
-	signals_notified[ss->index] &= ~(1 << ss->shift);
+	if (d->x_servers_order < 0)
+		d->x_servers_order = get_new_order (d);
 
-	return ((GDMSignalFunc)callback) (ss->signal, user_data);
-}
+	fp = fopen (file, "w");
+	if (fp == NULL) {
+		gdm_error ("Can't open %s for writing", file);
+		g_free (file);
+		return;
+	}
 
-static GSourceFuncs signal_funcs = {
-	gdm_signal_prepare,
-	gdm_signal_check,
-	gdm_signal_dispatch
-};
+	for (bogusname = 0, i = 0; i < d->x_servers_order; bogusname++, i++) {
+		char buf[32];
+		g_snprintf (buf, sizeof (buf), ":%d", bogusname);
+		if (strcmp (buf, d->name) == 0)
+			g_snprintf (buf, sizeof (buf), ":%d", ++bogusname);
+		fprintf (fp, "%s local /usr/X11R6/bin/Xbogus\n", buf);
+	}
 
-guint
-gdm_signal_add (gint8	      signal,
-		GDMSignalFunc function,
-		gpointer      data)
-{
-	return gdm_signal_add_full (G_PRIORITY_DEFAULT, signal, function, data, NULL);
-}
+	if (d->type == TYPE_XDMCP) {
+		fprintf (fp, "%s foreign\n", d->name);
+	} else {
+		char **argv;
+		char *command;
+		argv = gdm_server_resolve_command_line
+			(d, FALSE /* resolve_handled */);
+		command = g_strjoinv (" ", argv);
+		g_strfreev (argv);
+		fprintf (fp, "%s local %s\n", d->name, command);
+		g_free (command);
+	}
 
-guint
-gdm_signal_add_full (gint           priority,
-		     gint8          signal,
-		     GDMSignalFunc  function,
-		     gpointer       data,
-		     GDestroyNotify destroy)
-{
-	GSource *source;
-	SignalSource *ss;
-	guint s = 128 + signal;
-
-	g_return_val_if_fail (function != NULL, 0);
-
-	source = g_source_new (&signal_funcs, sizeof (SignalSource));
-	ss = (SignalSource *)source;
-
-	ss->signal = signal;
-	ss->index = s / 32;
-	ss->shift = s % 32;
-
-	g_source_set_priority (source, priority);
-	g_source_set_callback (source, (GSourceFunc)function, data, destroy);
-	g_source_set_can_recurse (source, TRUE);
-
-	return g_source_attach (source, NULL);
-}
-
-void
-gdm_signal_notify (gint8 signal)
-{
-	guint index, shift;
-	guint s = 128 + signal;
-
-	index = s / 32;
-	shift = s % 32;
-
-	signals_notified[index] |= 1 << shift;
+	fclose (fp);
 }
 
 static void
@@ -1649,6 +1644,22 @@ send_slave_ack (GdmDisplay *d)
 	}
 }
 
+static void
+send_slave_command (GdmDisplay *d, const char *command)
+{
+	if (d->master_notify_fd >= 0) {
+		char *cmd = g_strdup_printf ("%c%s\n",
+					     GDM_SLAVE_NOTIFY_COMMAND,
+					     command);
+		write (d->master_notify_fd, cmd, strlen (cmd));
+		g_free (cmd);
+	}
+	if (d->slavepid > 0) {
+		kill (d->slavepid, SIGUSR2);
+	}
+}
+
+
 static void
 gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 {
@@ -1661,7 +1672,11 @@ gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 			/* cut off most of the cookie for "security" */
 			gdm_debug ("Handling message: '%s...'", s);
 			g_free (s);
-		} else {
+		} else if (strncmp (msg, GDM_SOP_SYSLOG " ",
+				    strlen (GDM_SOP_SYSLOG " ")) != 0) {
+			/* Don't print out the syslog message as it will
+			 * be printed out anyway as that's the whole point
+			 * of the message. */
 			gdm_debug ("Handling message: '%s'", msg);
 		}
 	}
@@ -1933,6 +1948,18 @@ gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 	} else if (strcmp (msg, GDM_SOP_SOFT_RESTART) == 0) {
 		gdm_restart_mode = TRUE;
 		gdm_safe_restart ();
+	} else if (strcmp (msg, GDM_SOP_DIRTY_SERVERS) == 0) {
+		GSList *li;
+		for (li = displays; li != NULL; li = li->next) {
+			GdmDisplay *d = li->data;
+			send_slave_command (d, GDM_NOTIFY_DIRTY_SERVERS);
+		}
+	} else if (strcmp (msg, GDM_SOP_SOFT_RESTART_SERVERS) == 0) {
+		GSList *li;
+		for (li = displays; li != NULL; li = li->next) {
+			GdmDisplay *d = li->data;
+			send_slave_command (d, GDM_NOTIFY_SOFT_RESTART_SERVERS);
+		}
 	} else if (strncmp (msg, GDM_SOP_SYSLOG " ",
 		            strlen (GDM_SOP_SYSLOG " ")) == 0) {
 		char *p;
@@ -1967,6 +1994,24 @@ gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 			if (d->greetpid > 0)
 				kill (d->greetpid, SIGHUP);
 		}
+	} else if (strncmp (msg, GDM_SOP_WRITE_X_SERVERS " ",
+		            strlen (GDM_SOP_WRITE_X_SERVERS " ")) == 0) {
+		GdmDisplay *d;
+		long slave_pid;
+
+		if (sscanf (msg, GDM_SOP_WRITE_X_SERVERS " %ld",
+			    &slave_pid) != 1)
+			return;
+
+		/* Find out who this slave belongs to */
+		d = gdm_display_lookup (slave_pid);
+
+		if (d != NULL) {
+			write_x_servers (d);
+
+			/* send ack */
+			send_slave_ack (d);
+		}
 	}
 }
 
