@@ -17,26 +17,36 @@
  */
 
 #include "config.h"
-#include <gnome.h>
+#include <libgnome/libgnome.h>
 #include <syslog.h>
+#include <unistd.h>
 #include <signal.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/wait.h>
 #include <sys/socket.h>
+#include <sys/ioctl.h>
 #include <fcntl.h>
 #include <netinet/in.h>
+#include <net/if.h>
 #include <arpa/inet.h>
 #include <netdb.h>
 #include <errno.h>
+#include <grp.h>
+#include <sys/types.h>
+#ifdef HAVE_SYS_SOCKIO_H
+#include <sys/sockio.h>
+#endif
 
 #include <vicious.h>
 
 #include "gdm.h"
 #include "misc.h"
+#include "slave.h"
 
 /* Configuration option variables */
 extern gchar *GdmPidFile;
+extern gchar *GdmLocalNoPasswordUsers;
 extern gboolean GdmDebug;
 extern GSList *displays;
 extern int gdm_xdmcpfd;
@@ -45,6 +55,22 @@ extern char **environ;
 
 extern pid_t extra_process;
 extern int extra_status;
+extern pid_t gdm_main_pid;
+extern gboolean preserve_ld_vars;
+extern int gdm_in_signal;
+
+static void
+do_syslog (int type, const char *s)
+{
+	if (gdm_in_signal > 0) {
+		char *m = g_strdup_printf (GDM_SOP_SYSLOG " %ld %d %s",
+					   (long)getpid (), type, s);
+		gdm_slave_send (m, FALSE);
+		g_free (m);
+	} else {
+		syslog (type, "%s", s);
+	}
+}
 
 
 /**
@@ -60,21 +86,34 @@ void
 gdm_fail (const gchar *format, ...)
 {
     va_list args;
-    gchar *s;
+    char *s;
 
     va_start (args, format);
     s = g_strdup_vprintf (format, args);
     va_end (args);
 
     /* Log to both syslog and stderr */
-    syslog (LOG_CRIT, "%s", s);
-    fprintf (stderr, "%s\n", s);
-    fflush (stderr);
+    do_syslog (LOG_CRIT, s);
+    if (getpid () == gdm_main_pid) {
+	    gdm_fdprintf (2, "%s\n", s);
+    }
 
     g_free (s);
 
-    if (GdmPidFile != NULL)
-	    unlink (GdmPidFile);
+    /* If main process do final cleanup to kill all processes */
+    if (getpid () == gdm_main_pid) {
+	    gdm_final_cleanup ();
+    } else if ( ! gdm_slave_final_cleanup ()) {
+	    /* If we weren't even a slave do some random cleanup only */
+	    /* FIXME: is this all fine? */
+	    if (extra_process > 1 && extra_process != getpid ()) {
+		    /* we sigterm extra processes, and we
+		     * don't wait */
+		    kill (extra_process, SIGTERM);
+		    extra_process = -1;
+	    }
+    }
+
     closelog ();
 
     /* Slow down respawning if we're started from init */
@@ -97,13 +136,13 @@ void
 gdm_info (const gchar *format, ...)
 {
     va_list args;
-    gchar *s;
+    char *s;
 
     va_start (args, format);
     s = g_strdup_vprintf (format, args);
     va_end (args);
-    
-    syslog (LOG_INFO, "%s", s);
+
+    do_syslog (LOG_INFO, s);
     
     g_free (s);
 }
@@ -121,13 +160,13 @@ void
 gdm_error (const gchar *format, ...)
 {
     va_list args;
-    gchar *s;
+    char *s;
 
     va_start (args, format);
     s = g_strdup_vprintf (format, args);
     va_end (args);
-    
-    syslog (LOG_ERR, "%s", s);
+
+    do_syslog (LOG_ERR, s);
     
     g_free (s);
 }
@@ -145,26 +184,39 @@ void
 gdm_debug (const gchar *format, ...)
 {
     va_list args;
-    gchar *s;
+    char *s;
 
-    if (/*0 && */! GdmDebug) 
+    if ( ! GdmDebug) 
 	return;
 
     va_start (args, format);
     s = g_strdup_vprintf (format, args);
     va_end (args);
 
-    /* UGLY DEBUGGING HACK! */
-    
-    /*{ FILE *fp = fopen ("/tmp/foo.gdm", "a"); fprintf (fp, "%s\n", s); fflush (fp); fclose (fp); };*/
-    
-    syslog (LOG_ERR, "%s", s);	/* FIXME: LOG_DEBUG */
+    do_syslog (LOG_ERR, s);	/* Maybe should be LOG_DEBUG, but then normally
+				 * you wouldn't get it in the log.  ??? */
     
     g_free (s);
 }
 
+void 
+gdm_fdprintf (int fd, const gchar *format, ...)
+{
+	va_list args;
+	gchar *s;
+
+	va_start (args, format);
+	s = g_strdup_vprintf (format, args);
+	va_end (args);
+
+	write (fd, s, strlen (s));
+
+	g_free (s);
+}
+
 /* clear environment, but keep the i18n ones,
- * note that this leaks memory so only use before exec */
+ * note that this leaks memory so only use before exec
+ * (keep LD_* if preserve_ld_vars is true) */
 void
 gdm_clearenv_no_lang (void)
 {
@@ -177,9 +229,37 @@ gdm_clearenv_no_lang (void)
 		    strncmp (env, "LANG", 4) == 0 ||
 		    strncmp (env, "LINGUAS", 7) == 0)
 			envs = g_list_prepend (envs, g_strdup (env));
+		if (preserve_ld_vars &&
+		    strncmp (env, "LD_", 3) == 0)
+			envs = g_list_prepend (envs, g_strdup (env));
 	}
 
-	ve_clearenv ();
+	gnome_clearenv ();
+
+	for (li = envs; li != NULL; li = li->next) {
+		putenv (li->data);
+	}
+
+	g_list_free (envs);
+}
+
+/* clear environment completely
+ * note that this leaks memory so only use before exec
+ * (keep LD_* if preserve_ld_vars is true) */
+void
+gdm_clearenv (void)
+{
+	int i;
+	GList *li, *envs = NULL;
+
+	for (i = 0; environ[i] != NULL; i++) {
+		char *env = environ[i];
+		if (preserve_ld_vars &&
+		    strncmp (env, "LD_", 3) == 0)
+			envs = g_list_prepend (envs, g_strdup (env));
+	}
+
+	gnome_clearenv ();
 
 	for (li = envs; li != NULL; li = li->next) {
 		putenv (li->data);
@@ -263,11 +343,11 @@ gdm_text_message_dialog (const char *msg)
 	if (access (EXPANDED_SBINDIR "/gdmopen", X_OK) != 0)
 		return FALSE;
 
-	dialog = gnome_is_program_in_path ("dialog");
+	dialog = g_find_program_in_path ("dialog");
 	if (dialog == NULL)
-		dialog = gnome_is_program_in_path ("gdialog");
+		dialog = g_find_program_in_path ("gdialog");
 	if (dialog == NULL)
-		dialog = gnome_is_program_in_path ("whiptail");
+		dialog = g_find_program_in_path ("whiptail");
 	if (dialog != NULL) {
 		char *argv[7];
 
@@ -280,7 +360,8 @@ gdm_text_message_dialog (const char *msg)
 		argv[6] = NULL;
 
 		/* make sure gdialog wouldn't get confused */
-		if (gdm_exec_wait (argv, TRUE /* no display */) < 0) {
+		if (gdm_exec_wait (argv, TRUE /* no display */,
+				   TRUE /* de_setuid */) < 0) {
 			g_free (dialog);
 			return FALSE;
 		}
@@ -298,7 +379,8 @@ gdm_text_message_dialog (const char *msg)
 			 msg);
 		argv[4] = NULL;
 
-		if (gdm_exec_wait (argv, TRUE /* no display */) < 0) {
+		if (gdm_exec_wait (argv, TRUE /* no display */,
+				   TRUE /* de_setuid */) < 0) {
 			g_free (argv[3]);
 			return FALSE;
 		}
@@ -318,11 +400,11 @@ gdm_text_yesno_dialog (const char *msg, gboolean *ret)
 	if (ret != NULL)
 		*ret = FALSE;
 
-	dialog = gnome_is_program_in_path ("dialog");
+	dialog = g_find_program_in_path ("dialog");
 	if (dialog == NULL)
-		dialog = gnome_is_program_in_path ("gdialog");
+		dialog = g_find_program_in_path ("gdialog");
 	if (dialog == NULL)
-		dialog = gnome_is_program_in_path ("whiptail");
+		dialog = g_find_program_in_path ("whiptail");
 	if (dialog != NULL) {
 		char *argv[7];
 		int retint;
@@ -337,7 +419,8 @@ gdm_text_yesno_dialog (const char *msg, gboolean *ret)
 
 		/* will unset DISPLAY and XAUTHORITY if they exist
 		 * so that gdialog (if used) doesn't get confused */
-		retint = gdm_exec_wait (argv, TRUE /* no display */);
+		retint = gdm_exec_wait (argv, TRUE /* no display */,
+					TRUE /* de_setuid */);
 		if (retint < 0) {
 			g_free (dialog);
 			return FALSE;
@@ -356,7 +439,7 @@ gdm_text_yesno_dialog (const char *msg, gboolean *ret)
 		char buf[256];
 		char *argv[5];
 
-		tempfd = mkstemp (tempname);
+		tempfd = g_mkstemp (tempname);
 		if (tempfd < 0)
 			return FALSE;
 
@@ -375,7 +458,8 @@ gdm_text_yesno_dialog (const char *msg, gboolean *ret)
 			 tempname);
 		argv[4] = NULL;
 
-		if (gdm_exec_wait (argv, TRUE /* no display */) < 0) {
+		if (gdm_exec_wait (argv, TRUE /* no display */,
+				   TRUE /* de_setuid */) < 0) {
 			g_free (argv[3]);
 			return FALSE;
 		}
@@ -400,7 +484,8 @@ gdm_text_yesno_dialog (const char *msg, gboolean *ret)
 }
 
 int
-gdm_exec_wait (char * const *argv, gboolean no_display)
+gdm_exec_wait (char * const *argv, gboolean no_display,
+	       gboolean de_setuid)
 {
 	int status;
 	pid_t pid;
@@ -412,20 +497,25 @@ gdm_exec_wait (char * const *argv, gboolean no_display)
 
 	pid = gdm_fork_extra ();
 	if (pid == 0) {
-		int i;
+		closelog ();
 
-		for (i = 0; i < sysconf (_SC_OPEN_MAX); i++)
-			close (i);
+		gdm_close_all_descriptors (0 /* from */, -1 /* except */);
 
 		/* No error checking here - if it's messed the best response
 		 * is to ignore & try to continue */
-		open ("/dev/null", O_RDONLY); /* open stdin - fd 0 */
-		open ("/dev/null", O_RDWR); /* open stdout - fd 1 */
-		open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
+		gdm_open_dev_null (O_RDONLY); /* open stdin - fd 0 */
+		gdm_open_dev_null (O_RDWR); /* open stdout - fd 1 */
+		gdm_open_dev_null (O_RDWR); /* open stderr - fd 2 */
+
+		if (de_setuid) {
+			gdm_desetuid ();
+		}
+
+		openlog ("gdm", LOG_PID, LOG_DAEMON);
 
 		if (no_display) {
-			ve_unsetenv ("DISPLAY");
-			ve_unsetenv ("XAUTHORITY");
+			gnome_unsetenv ("DISPLAY");
+			gnome_unsetenv ("XAUTHORITY");
 		}
 		
 		execv (argv[0], argv);
@@ -450,6 +540,9 @@ static sigset_t sigchldblock_mask, sigchldblock_oldmask;
 static int sigterm_blocked = 0;
 static sigset_t sigtermblock_mask, sigtermblock_oldmask;
 
+static int sigusr2_blocked = 0;
+static sigset_t sigusr2block_mask, sigusr2block_oldmask;
+
 void
 gdm_sigchld_block_push (void)
 {
@@ -500,6 +593,30 @@ gdm_sigterm_block_pop (void)
 	}
 }
 
+void
+gdm_sigusr2_block_push (void)
+{
+	sigusr2_blocked ++;
+
+	if (sigusr2_blocked == 1) {
+		/* Set signal mask */
+		sigemptyset (&sigusr2block_mask);
+		sigaddset (&sigusr2block_mask, SIGUSR2);
+		sigprocmask (SIG_BLOCK, &sigusr2block_mask, &sigusr2block_oldmask);
+	}
+}
+
+void
+gdm_sigusr2_block_pop (void)
+{
+	sigusr2_blocked --;
+
+	if (sigusr2_blocked == 0) {
+		/* reset signal mask back */
+		sigprocmask (SIG_SETMASK, &sigusr2block_oldmask, NULL);
+	}
+}
+
 pid_t
 gdm_fork_extra (void)
 {
@@ -521,7 +638,7 @@ gdm_wait_for_extra (int *status)
 	gdm_sigchld_block_push ();
 
 	if (extra_process > 0) {
-		waitpid (extra_process, &extra_status, 0);
+		ve_waitpid_no_signal (extra_process, &extra_status, 0);
 	}
 	extra_process = -1;
 
@@ -606,14 +723,16 @@ gdm_peek_local_address_list (void)
 #endif
 
 	ifc.ifc_len = sizeof(struct ifreq) * num;
-	ifc.ifc_buf = buf = g_malloc (ifc.ifc_len);
+	ifc.ifc_buf = buf = g_malloc0 (ifc.ifc_len);
 	if (ioctl (sockfd, SIOCGIFCONF, &ifc) < 0) {
 		gdm_error (_("%s: Cannot get local addresses!"),
 			   "gdm_peek_local_address_list");
 		g_free (buf);
 		if (sockfd != gdm_xdmcpfd)
 			close (sockfd);
-		return NULL;
+		addr = g_new0 (struct in_addr, 1);
+		addr->s_addr = INADDR_LOOPBACK;
+		return g_list_append (NULL, addr);
 	}
 
 	ifr = ifc.ifc_req;
@@ -648,17 +767,22 @@ gdm_peek_local_address_list (void)
 #else /* SIOCGIFCONF */
 	/* host based fallback, will likely only get 127.0.0.1 i think */
 
+	hostbuf[BUFSIZ-1] = '\0';
 	if (gethostname (hostbuf, BUFSIZ-1) != 0) {
-		gdm_error (_("%s: Could not get server hostname: %s!"),
+		gdm_debug ("%s: Could not get server hostname: %s!",
 			   "gdm_peek_local_address_list",
 			   g_strerror (errno));
-		return NULL;
+		addr = g_new0 (struct in_addr, 1);
+		addr->s_addr = INADDR_LOOPBACK;
+		return g_list_append (NULL, addr);
 	}
 	he = gethostbyname (hostbuf);
 	if (he == NULL) {
-		gdm_error (_("%s: Could not get address from hostname!"),
+		gdm_debug ("%s: Could not get address from hostname!",
 			   "gdm_peek_local_address_list");
-		return NULL;
+		addr = g_new0 (struct in_addr, 1);
+		addr->s_addr = INADDR_LOOPBACK;
+		return g_list_append (NULL, addr);
 	}
 	for (i = 0; he->h_addr_list[i] != NULL; i++) {
 		struct in_addr *laddr = (struct in_addr *)he->h_addr_list[i];
@@ -710,5 +834,198 @@ gdm_is_loopback_addr (struct in_addr *ia)
 	}
 }
 
+gboolean
+gdm_setup_gids (const char *login, gid_t gid)
+{
+	/* FIXME: perhaps for *BSD there should be setusercontext
+	 * stuff here */
+	if (setgid (gid) < 0)  {
+		gdm_error (_("Could not setgid %d. Aborting."), (int)gid);
+		return FALSE;
+	}
+
+	if (initgroups (login, gid) < 0) {
+		gdm_error (_("initgroups() failed for %s. Aborting."), login);
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
+void
+gdm_desetuid (void)
+{
+	uid_t uid = getuid (); 
+	gid_t gid = getgid (); 
+
+#ifdef HAVE_SETRESUID
+	{
+		int setresuid(uid_t ruid, uid_t euid, uid_t suid);
+		int setresgid(gid_t rgid, gid_t egid, gid_t sgid);
+		setresuid (uid, uid, uid);
+		setresgid (gid, gid, gid);
+	}
+#else
+	seteuid (getuid ());
+	setegid (getgid ());
+#endif
+}
+
+gboolean
+gdm_test_opt (const char *cmd, const char *help, const char *option)
+{
+	char *q;
+	char *full;
+	char buf[1024];
+	FILE *fp;
+	static GString *cache = NULL;
+	static char *cached_cmd = NULL;
+	gboolean got_it;
+
+	if (cached_cmd != NULL &&
+	    strcmp (cached_cmd, cmd) == 0) {
+		char *p = strstr (ve_sure_string (cache->str), option);
+		char end;
+		if (p == NULL)
+			return FALSE;
+		/* must be a full word */
+		end = *(p + strlen (option));
+		if ((end >= 'a' && end <= 'z') ||
+		    (end >= 'A' && end <= 'Z') ||
+		    (end >= '0' && end <= '9') ||
+		    end == '_')
+			return FALSE;
+		return TRUE;
+	}
+
+	g_free (cached_cmd);
+	cached_cmd = g_strdup (cmd);
+	if (cache != NULL)
+		g_string_assign (cache, "");
+	else
+		cache = g_string_new (NULL);
+
+	q = g_shell_quote (cmd);
+
+	full = g_strdup_printf ("%s %s 2>&1", q, help);
+	g_free (q);
+
+	fp = popen (full, "r");
+	g_free (full);
+
+	if (fp == NULL)
+		return FALSE;
+
+	got_it = FALSE;
+
+	while (fgets (buf, sizeof (buf), fp) != NULL) {
+		char *p;
+		char end;
+
+		g_string_append (cache, buf);
+
+		if (got_it)
+			continue;
+
+		p = strstr (buf, option);
+		if (p == NULL)
+			continue;
+		/* must be a full word */
+		end = *(p + strlen (option));
+		if ((end >= 'a' && end <= 'z') ||
+		    (end >= 'A' && end <= 'Z') ||
+		    (end >= '0' && end <= '9') ||
+		    end == '_')
+			continue;
+
+		got_it = TRUE;
+	}
+	fclose (fp);
+	return got_it;
+}
+
+int
+gdm_fdgetc (int fd)
+{
+	char buf[1];
+	int bytes;
+
+	bytes = read (fd, buf, 1);
+	if (bytes != 1)
+		return EOF;
+	else
+		return (int)buf[0];
+}
+
+char *
+gdm_fdgets (int fd)
+{
+	int c;
+	int bytes = 0;
+	GString *gs = g_string_new (NULL);
+	for (;;) {
+		c = gdm_fdgetc (fd);
+		if (c == '\n')
+			return g_string_free (gs, FALSE);
+		/* on EOF */
+		if (c < 0) {
+			if (bytes == 0) {
+				g_string_free (gs, TRUE);
+				return NULL;
+			} else {
+				return g_string_free (gs, FALSE);
+			}
+		} else {
+			bytes ++;
+			g_string_append_c (gs, c);
+		}
+	}
+}
+
+void
+gdm_close_all_descriptors (int from, int except)
+{
+	int i;
+	int max = sysconf (_SC_OPEN_MAX);
+	for (i = from; i < max; i++) {
+		if (i != except)
+			close(i);
+	}
+}
+
+gboolean
+gdm_is_a_no_password_user (const char *user)
+{
+	char **vector = NULL;
+	int i;
+
+	if (ve_string_empty (GdmLocalNoPasswordUsers) ||
+	    ve_string_empty (user) ||
+	    strcmp (user, "root") == 0)
+		return FALSE;
+
+	vector = g_strsplit (GdmLocalNoPasswordUsers, ",", -1);
+	for (i = 0; vector[i] != NULL; i++) {
+		if (strcmp (vector[i], user) == 0) {
+			g_strfreev (vector);
+			return TRUE;
+		}
+	}
+	g_strfreev (vector);
+	return FALSE;
+}
+
+int
+gdm_open_dev_null (mode_t mode)
+{
+	int ret;
+	ret = open ("/dev/null", mode);
+	if (ret < 0) {
+		gdm_fail ("Cannot open /dev/null, system on crack!");
+	}
+
+	return ret;
+}
+
 
 /* EOF */
