@@ -21,16 +21,23 @@
 #ifndef VE_MISC_H
 #define VE_MISC_H
 
+#include <glib.h>
 #include <time.h>
 #include <sys/types.h>
 #include <sys/wait.h>
 
 char **	ve_split (const char *s);
 char *	ve_first_word (const char *s);
+gboolean ve_first_word_executable (const char *s,
+				   gboolean only_existance);
 char *  ve_rest (const char *s);
 char **	ve_vector_merge (char * const *v1, char * const *v2);
 int	ve_vector_len (char * const *v);
 
+/* Gets the first existing command out of a list separated by semicolons */
+char *	ve_get_first_working_command (const char *list,
+				      gboolean only_existance);
+
 gboolean ve_bool_equal (gboolean a, gboolean b);
 
 gboolean ve_is_string_in_list (const GList *list,
@@ -41,6 +48,12 @@ gboolean ve_is_string_in_list_case_no_locale (const GList *list,
 #define ve_string_empty(x) ((x)==NULL||(x)[0]=='\0')
 #define ve_sure_string(x) ((x)!=NULL?(x):"")
 
+/* Find a file using the specified list of directories, an absolute path is
+ * just checked, whereas a relative path is search in the given directories,
+ */
+char *		ve_find_file_simple		(const char *filename,
+						 const GList *directories);
+
 /* Find a file using the specified list of directories, an absolute path is
  * just checked, whereas a relative path is search in the given directories,
  * gnome_datadir, g_get_prgname() subdirectory, in GNOME_PATH
